<?php

namespace FiloBlu\AdminSecurity\Model;

use Magento\Backend\Model\Auth\Session;
use Magento\Framework\App\RequestInterface;
use Magento\Framework\HTTP\PhpEnvironment\RemoteAddress;
use Magento\Framework\Json\Decoder;
use Magento\Framework\Json\Encoder;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\User\Model\User;
use FiloBlu\AdminSecurity\Api\TrustedManagerInterface;
use FiloBlu\AdminSecurity\Api\TrustedRepositoryInterface;
use FiloBlu\AdminSecurity\Model\ResourceModel\Trusted as TrustedResourceModel;
use Magento\Framework\Stdlib\CookieManagerInterface;
use Magento\Framework\Session\SessionManagerInterface;
use Magento\Framework\Stdlib\Cookie\CookieMetadataFactory;

/**
 * Class TrustedManager
 * @package FiloBlu\AdminSecurity\Model
 * @SuppressWarnings("PHPMD.CouplingBetweenObjects")
 */
class TrustedManager implements TrustedManagerInterface
{
    private $isTrustedDevice = null;

    /**
     * @var TrustedFactory
     */
    private $trustedFactory;

    /**
     * @var DateTime
     */
    private $dateTime;

    /**
     * @var RemoteAddress
     */
    private $remoteAddress;

    /**
     * @var Session
     */
    private $session;

    /**
     * @var TrustedResourceModel
     */
    private $trustedResourceModel;

    /**
     * @var CookieManagerInterface
     */
    private $cookieManager;

    /**
     * @var SessionManagerInterface
     */
    private $sessionManager;

    /**
     * @var CookieMetadataFactory
     */
    private $cookieMetadataFactory;

    /**
     * @var Encoder
     */
    private $encoder;

    /**
     * @var Decoder
     */
    private $decoder;
    /**
     * @var TrustedRepositoryInterface
     */
    private $trustedRepository;

    /**
     * TrustedManager constructor.
     * @param DateTime $dateTime
     * @param Session $session
     * @param RemoteAddress $remoteAddress
     * @param Encoder $encoder
     * @param Decoder $decoder
     * @param TrustedResourceModel $trustedResourceModel
     * @param CookieManagerInterface $cookieManager
     * @param SessionManagerInterface $sessionManager
     * @param TrustedRepositoryInterface $trustedRepository
     * @param TrustedFactory $trustedFactory
     * @param CookieMetadataFactory $cookieMdFactory
     * @SuppressWarnings("PHPMD.ExcessiveParameterList")
     */
    public function __construct(
        DateTime $dateTime,
        Session $session,
        RemoteAddress $remoteAddress,
        Encoder $encoder,
        Decoder $decoder,
        TrustedResourceModel $trustedResourceModel,
        CookieManagerInterface $cookieManager,
        SessionManagerInterface $sessionManager,
        TrustedRepositoryInterface $trustedRepository,
        TrustedFactory $trustedFactory,
        CookieMetadataFactory $cookieMdFactory
    ) {
        $this->trustedFactory = $trustedFactory;
        $this->dateTime = $dateTime;
        $this->remoteAddress = $remoteAddress;
        $this->session = $session;
        $this->trustedResourceModel = $trustedResourceModel;
        $this->cookieManager = $cookieManager;
        $this->sessionManager = $sessionManager;
        $this->cookieMetadataFactory = $cookieMdFactory;
        $this->encoder = $encoder;
        $this->decoder = $decoder;
        $this->trustedRepository = $trustedRepository;
    }

    /**
     * Get current user
     * @return User|null
     */
    private function getUser()
    {
        return $this->session->getUser();
    }

    /**
     * Get device name
     * @return string
     */
    private function getDeviceName()
    {
        $browser = parse_user_agent();
        return $browser['platform'] . ' ' . $browser['browser'] . ' ' . $browser['version'];
    }

    /**
     * Get token collection from cookie
     * @return array
     */
    private function getTokenCollection()
    {
        try {
            return $this->decoder->decode(
                $this->cookieManager->getCookie(TrustedManagerInterface::TRUSTED_DEVICE_COOKIE)
            );
        } catch (\Exception $e) {
            return [];
        }
    }

    /**
     * Send token as cookie
     * @param string $token
     * @param User $user
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Stdlib\Cookie\CookieSizeLimitReachedException
     * @throws \Magento\Framework\Stdlib\Cookie\FailureToSendException
     */
    private function sendTokenCookie($token, User $user)
    {
        $tokenCollection = $this->getTokenCollection();

        // Enable cookie
        $cookieMetadata = $this->cookieMetadataFactory->createPublicCookieMetadata()
            ->setDurationOneYear()
            ->setHttpOnly(true)
            ->setPath($this->sessionManager->getCookiePath())
            ->setDomain($this->sessionManager->getCookieDomain());

        $tokenCollection[$user->getUserName()] = $token;

        $this->cookieManager->setPublicCookie(
            TrustedManagerInterface::TRUSTED_DEVICE_COOKIE,
            $this->encoder->encode($tokenCollection),
            $cookieMetadata
        );
    }

    /**
     * Rotate secret trust token
     * @param User $user
     * @return void
     * @throws \Exception
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Stdlib\Cookie\CookieSizeLimitReachedException
     * @throws \Magento\Framework\Stdlib\Cookie\FailureToSendException
     */
    public function rotateTrustedDeviceToken(User $user)
    {
        $tokenCollection = $this->getTokenCollection();

        if (isset($tokenCollection[$user->getUserName()])) {
            $token = $tokenCollection[$user->getUserName()];

            /** @var $trustEntry Trusted */
            $trustEntry = $this->trustedFactory->create();
            $this->trustedResourceModel->load($trustEntry, $token, 'token');
            if ($trustEntry->getId() && ($trustEntry->getUserId() == $user->getId())) {
                $token = sha1(uniqid(time()));

                $trustEntry->setToken($token);
                $this->trustedResourceModel->save($trustEntry);

                $this->sendTokenCookie($token, $user);
            }
        }
    }

    /**
     * Return true if device is trusted
     * @param User $user
     * @return bool
     */
    public function isTrustedDevice(User $user)
    {
        if ($this->isTrustedDevice === null) { // Must cache this ina single session to avoid rotation issues
            $tokenCollection = $this->getTokenCollection();

            if (isset($tokenCollection[$user->getUserName()])) {
                $token = $tokenCollection[$user->getUserName()];

                /** @var $trustEntry Trusted */
                $trustEntry = $this->trustedFactory->create();
                $this->trustedResourceModel->load($trustEntry, $token, 'token');

                $this->isTrustedDevice = $trustEntry->getId() && ($trustEntry->getUserId() == $user->getId());
            } else {
                $this->isTrustedDevice = false;
            }
        }

        return $this->isTrustedDevice;
    }

    /**
     * Revoke trusted device
     * @param int $tokenId
     * @return bool
     */
    public function revokeTrustedDevice($tokenId)
    {
        $token = $this->trustedRepository->getById($tokenId);
        $this->trustedRepository->delete($token);

        return true;
    }

    /**
     * Trust a device
     * @param RequestInterface $request
     * @param User $user
     * @return boolean
     * @throws \Exception
     * @throws \Magento\Framework\Exception\InputException
     * @throws \Magento\Framework\Stdlib\Cookie\CookieSizeLimitReachedException
     * @throws \Magento\Framework\Stdlib\Cookie\FailureToSendException
     */
    public function handleTrustDeviceRequest(RequestInterface $request, User $user)
    {
        if ($request->getParam('filoblu_as_trust_device') &&
            ($request->getParam('filoblu_as_trust_device') != "false") // u2fkey submit translates into a string
        ) {
            $token = sha1(uniqid(time()));

            /** @var $trustEntry Trusted */
            $trustEntry = $this->trustedFactory->create();
            $trustEntry
                ->setToken($token)
                ->setDateTime($this->dateTime->date())
                ->setUserId($user->getId())
                ->setLastIp($this->remoteAddress->getRemoteAddress())
                ->setDeviceName($this->getDeviceName())
                ->setUserAgent($request->getServer('HTTP_USER_AGENT'));

            $this->trustedResourceModel->save($trustEntry);

            $this->sendTokenCookie($token, $user);
            return true;
        }

        return false;
    }
}
